# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import io
import os

from pyomo.contrib.cspline_external.cspline_parameters import (
    cubic_parameters_model,
    CsplineParameters,
)
from pyomo.opt import check_available_solvers
from pyomo.common.dependencies import numpy as np, numpy_available
import pyomo.environ as pyo
from pyomo.common.fileutils import this_file_dir
import pyomo.common.unittest as unittest


@unittest.skipUnless(
    check_available_solvers("ipopt"), "The 'ipopt' solver is not available"
)
@unittest.skipIf(not numpy_available, "numpy is not available.")
class CsplineExternalParamsTest(unittest.TestCase):
    def test_param_gen(self):
        x_data = [1, 2, 3, 4, 5]
        y_data = [2, 3, 5, 2, 1]

        m = cubic_parameters_model(x_data, y_data)

        # This is a linear system of equations by default, so although ipopt
        # is not needed, it works perfectly well.
        solver_obj = pyo.SolverFactory("ipopt")
        solver_obj.solve(m)

        params = CsplineParameters(model=m)

        # Make sure we have the expected number of parameters
        assert params.n_knots == 5
        assert params.n_segments == 4
        assert params.valid

        # Make sure the predictions are correct
        y_pred = params.f(np.array(x_data))
        for yd, yp in zip(y_data, y_pred):
            self.assertAlmostEqual(yd, yp, 8)

        # Make sure reading and writing parameters
        # to/from files works right.

        fptr = io.StringIO("")
        params.write_parameters(fptr)
        # Go to start of string stream
        fptr.seek(0, os.SEEK_SET)

        params.knots = np.array([1, 3])
        assert not params.valid
        params.get_parameters_from_file(fptr)
        assert params.valid

        # Make sure the predictions are correct
        y_pred = params.f(np.array(x_data))
        for yd, yp in zip(y_data, y_pred):
            self.assertAlmostEqual(yd, yp, 8)

        # Make sure the predictions are correct
        y_pred = params.f(np.array(x_data) + 1e-4)
        for yd, yp in zip(y_data, y_pred):
            self.assertAlmostEqual(yd, yp, 3)

        # Make sure the predictions are correct
        y_pred = params.f(np.array(x_data) - 1e-4)
        for yd, yp in zip(y_data, y_pred):
            self.assertAlmostEqual(yd, yp, 3)

        self.assertAlmostEqual(1, params.f(5), 8)
        self.assertAlmostEqual(1, params.f(5 - 1e-4), 3)
        self.assertAlmostEqual(1, params.f(5 + 1e-4), 3)
        self.assertAlmostEqual(2, params.f(1), 8)
        self.assertAlmostEqual(2, params.f(1 - 1e-4), 3)
        self.assertAlmostEqual(2, params.f(1 + 1e-4), 3)

    def test_param_increasing(self):
        x_data = [1, 2, 3, 4, 5]
        y_data = [1, 4, 9, 16, 25]

        m = cubic_parameters_model(x_data, y_data, objective_form=True)
        m.add_increasing_constraints()
        solver_obj = pyo.SolverFactory("ipopt")
        solver_obj.solve(m)

        params = CsplineParameters(model=m)

        # Make sure the predictions are correct
        y_pred = params.f(np.array(x_data))
        for yd, yp in zip(y_data, y_pred):
            self.assertAlmostEqual(yd, yp, 6)

        x_new = [1.5, 2.5, 3.5, 4.5]
        y_new = [2.25, 6.25, 12.25, 20.25]
        # Check midpoints.  These aren't necessarily super close due
        # to the end point second derivative constraints, and I
        # calculated test values from a quadratic.
        y_pred = params.f(np.array(x_new))
        for yd, yp in zip(y_new, y_pred):
            self.assertAlmostEqual(yd, yp, 0)

    def test_param_increasing_linear_extrap(self):
        x_data = [1, 2, 3, 4, 5]
        y_data = [1, 4, 9, 16, 25]

        m = cubic_parameters_model(x_data, y_data, objective_form=True)
        m.add_increasing_constraints()

        solver_obj = pyo.SolverFactory("ipopt")
        solver_obj.solve(m)

        params = CsplineParameters(model=m)
        params.add_linear_extrapolation_segments()

        # Make sure the predictions are correct
        y_pred = params.f(np.array(x_data))
        for yd, yp in zip(y_data, y_pred):
            self.assertAlmostEqual(yd, yp, 6)

        x_new = [0, 6, 7, 8]
        y_new = [-1.571, 34.42857, 43.8571, 53.2857]
        y_pred = params.f(np.array(x_new))
        for yd, yp in zip(y_new, y_pred):
            self.assertAlmostEqual(yd, yp, 3)

    def test_param_decreasing(self):
        x_data = [-1, -2, -3, -4, -5]
        y_data = [1, 4, 9, 16, 25]

        m = cubic_parameters_model(x_data, y_data, objective_form=True)
        m.add_decreasing_constraints()

        solver_obj = pyo.SolverFactory("ipopt")
        solver_obj.solve(m)

        params = CsplineParameters(model=m)

        # Make sure the predictions are correct
        y_pred = params.f(np.array(x_data))
        for yd, yp in zip(y_data, y_pred):
            self.assertAlmostEqual(yd, yp, 5)

        x_new = [-1.5, -2.5, -3.5, -4.5]
        y_new = [2.25, 6.25, 12.25, 20.25]
        # Check midpoints.  These aren't necessarily super close due
        # to the end point second derivative constraints, and I
        # calculated test values from a quadratic.
        y_pred = params.f(np.array(x_new))
        for yd, yp in zip(y_new, y_pred):
            self.assertAlmostEqual(yd, yp, 1)

    def test_convex(self):
        x_data = [-1, 1, 2, 3, 4, 5, 6]
        y_data = [1, 1, 4, 11, 16, 25, 36]

        m = cubic_parameters_model(
            x_data, y_data, objective_form=True, end_point_constraint=False
        )
        m.add_convex_constraints(tol=0.1)
        solver_obj = pyo.SolverFactory("ipopt")
        solver_obj.solve(m)
        params = CsplineParameters(model=m)
        y_pred = params.f(np.array(x_data))
        y_new = [1.007, 0.891, 4.524, 10.154, 16.536, 24.866, 36.0223]
        for yd, yp in zip(y_new, y_pred):
            self.assertAlmostEqual(yd, yp, 2)

    def test_concave(self):
        x_data = [1, 1, 2, 3, 4, 5, 6]
        y_data = [-1, -1, -4, -14, -16, -25, -36]
        m = cubic_parameters_model(
            x_data, y_data, objective_form=True, end_point_constraint=False
        )
        m.add_concave_constraints(tol=0.1)
        solver_obj = pyo.SolverFactory("ipopt")
        solver_obj.solve(m)
        params = CsplineParameters(model=m)
        y_pred = params.f(np.array(x_data))
        y_new = [-1.000, -1.000, -5.259, -11.3533, -17.5475, -24.80776, -36.032]
        for yd, yp in zip(y_new, y_pred):
            self.assertAlmostEqual(yd, yp, 2)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
from pyomo.common.fileutils import this_file_dir
from pyomo.contrib.simplification import Simplifier
from pyomo.contrib.simplification.simplify import ginac_available
from pyomo.core.expr.compare import assertExpressionsEqual, compare_expressions
from pyomo.core.expr.calculus.diff_with_pyomo import reverse_sd
from pyomo.core.expr.sympy_tools import sympy_available

import pyomo.environ as pyo


class SimplificationMixin:
    def compare_against_possible_results(self, got, expected_list):
        success = False
        for exp in expected_list:
            if compare_expressions(got, exp):
                success = True
                break
        self.assertTrue(success)

    def test_simplify(self):
        m = pyo.ConcreteModel()
        x = m.x = pyo.Var(bounds=(0, None))
        e = x * pyo.log(x)
        der1 = reverse_sd(e)[x]
        der2 = reverse_sd(der1)[x]
        der2_simp = self.simp.simplify(der2)
        expected = x**-1.0
        assertExpressionsEqual(self, expected, der2_simp)

    def test_mul(self):
        m = pyo.ConcreteModel()
        x = m.x = pyo.Var()
        e = 2 * x
        e2 = self.simp.simplify(e)
        expected = 2.0 * x
        assertExpressionsEqual(self, expected, e2)

    def test_sum(self):
        m = pyo.ConcreteModel()
        x = m.x = pyo.Var()
        e = 2 + x
        e2 = self.simp.simplify(e)
        self.compare_against_possible_results(e2, [2.0 + x, x + 2.0])

    def test_neg(self):
        m = pyo.ConcreteModel()
        x = m.x = pyo.Var()
        e = -pyo.log(x)
        e2 = self.simp.simplify(e)
        self.compare_against_possible_results(
            e2, [(-1.0) * pyo.log(x), pyo.log(x) * (-1.0), -pyo.log(x)]
        )

    def test_pow(self):
        m = pyo.ConcreteModel()
        x = m.x = pyo.Var()
        e = x**2.0
        e2 = self.simp.simplify(e)
        assertExpressionsEqual(self, e, e2)

    def test_div(self):
        m = pyo.ConcreteModel()
        x = m.x = pyo.Var()
        y = m.y = pyo.Var()
        e = x / y + y / x - x / y
        e2 = self.simp.simplify(e)
        self.compare_against_possible_results(
            e2, [y / x, y * (1.0 / x), y * x**-1.0, x**-1.0 * y]
        )

    def test_unary(self):
        m = pyo.ConcreteModel()
        x = m.x = pyo.Var()
        func_list = [pyo.log, pyo.sin, pyo.cos, pyo.tan, pyo.asin, pyo.acos, pyo.atan]
        for func in func_list:
            e = func(x)
            e2 = self.simp.simplify(e)
            assertExpressionsEqual(self, e, e2)

    def test_param(self):
        m = pyo.ConcreteModel()
        x = m.x = pyo.Var()
        p = m.p = pyo.Param(mutable=True)
        e1 = p * x**2 + p * x + p * x**2
        e2 = self.simp.simplify(e1)
        self.compare_against_possible_results(
            e2,
            [
                p * x**2.0 * 2.0 + p * x,
                p * x + p * x**2.0 * 2.0,
                2.0 * p * x**2.0 + p * x,
                p * x + 2.0 * p * x**2.0,
                x**2.0 * p * 2.0 + p * x,
                p * x + x**2.0 * p * 2.0,
                p * x * (1 + 2 * x),
            ],
        )


@unittest.skipUnless(sympy_available, 'sympy is not available')
class TestSimplificationSympy(unittest.TestCase, SimplificationMixin):
    def setUp(self):
        self.simp = Simplifier(mode=Simplifier.Mode.sympy)


@unittest.pytest.mark.default
@unittest.pytest.mark.builders
@unittest.skipUnless(ginac_available, 'GiNaC is not available')
class TestSimplificationGiNaC(unittest.TestCase, SimplificationMixin):
    def setUp(self):
        self.simp = Simplifier(mode=Simplifier.Mode.ginac)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.common.collections import ComponentSet
from pyomo.core.expr.compare import compare_expressions
from pyomo.contrib.mpc.interfaces.model_interface import DynamicModelInterface
from pyomo.contrib.mpc.data.series_data import TimeSeriesData
from pyomo.contrib.mpc.data.scalar_data import ScalarData
from pyomo.contrib.mpc.data.interval_data import IntervalData


class TestDynamicModelInterface(unittest.TestCase):
    def _make_model(self, n_time_points=3):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=range(n_time_points))
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(
            m.time,
            m.comp,
            initialize={(i, j): 1.0 + i * 0.1 for i, j in m.time * m.comp},
        )
        m.input = pyo.Var(m.time, initialize={i: 1.0 - i * 0.1 for i in m.time})
        m.scalar = pyo.Var(initialize=0.5)
        m.var_squared = pyo.Expression(
            m.time,
            m.comp,
            initialize={(i, j): m.var[i, j] ** 2 for i, j in m.time * m.comp},
        )
        return m

    def _hashRef(self, reference):
        return tuple(id(obj) for obj in reference.values())

    def test_interface_construct(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)

        scalar_vars = interface.get_scalar_variables()
        self.assertEqual(len(scalar_vars), 1)
        self.assertIs(scalar_vars[0], m.scalar)

        dae_vars = interface.get_indexed_variables()
        self.assertEqual(len(dae_vars), 3)
        dae_var_set = set(self._hashRef(var) for var in dae_vars)
        pred_dae_var = [
            pyo.Reference(m.var[:, "A"]),
            pyo.Reference(m.var[:, "B"]),
            m.input,
        ]
        for var in pred_dae_var:
            self.assertIn(self._hashRef(var), dae_var_set)

        dae_expr = interface.get_indexed_expressions()
        dae_expr_set = set(self._hashRef(expr) for expr in dae_expr)
        self.assertEqual(len(dae_expr), 2)
        pred_dae_expr = [
            pyo.Reference(m.var_squared[:, "A"]),
            pyo.Reference(m.var_squared[:, "B"]),
        ]
        for expr in pred_dae_expr:
            self.assertIn(self._hashRef(expr), dae_expr_set)

    def test_get_scalar_var_data(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = interface.get_scalar_variable_data()
        self.assertEqual({pyo.ComponentUID(m.scalar): 0.5}, data)

    def test_get_data_at_time_all_points(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = interface.get_data_at_time(include_expr=True)
        pred_data = TimeSeriesData(
            {
                m.var[:, "A"]: [1.0, 1.1, 1.2],
                m.var[:, "B"]: [1.0, 1.1, 1.2],
                m.input[:]: [1.0, 0.9, 0.8],
                m.var_squared[:, "A"]: [1.0**2, 1.1**2, 1.2**2],
                m.var_squared[:, "B"]: [1.0**2, 1.1**2, 1.2**2],
            },
            m.time,
        )
        self.assertEqual(data, pred_data)

    def test_get_data_at_time_subset(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = interface.get_data_at_time(time=[0, 2])
        pred_data = TimeSeriesData(
            {
                m.var[:, "A"]: [1.0, 1.2],
                m.var[:, "B"]: [1.0, 1.2],
                m.input[:]: [1.0, 0.8],
            },
            [0, 2],
        )
        self.assertEqual(data, pred_data)

    def test_get_data_at_time_singleton(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = interface.get_data_at_time(time=1, include_expr=True)
        pred_data = ScalarData(
            {
                m.var[:, "A"]: 1.1,
                m.var[:, "B"]: 1.1,
                m.input[:]: 0.9,
                m.var_squared[:, "A"]: 1.1**2,
                m.var_squared[:, "B"]: 1.1**2,
            }
        )
        self.assertEqual(data, pred_data)

    def test_load_scalar_data(self):
        # load_scalar_data has been removed. Instead we simply call
        # load_data
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = {pyo.ComponentUID(m.scalar): 6.0}
        interface.load_data(data)
        self.assertEqual(m.scalar.value, 6.0)

    def test_load_data_at_time_all(self):
        # NOTE: load_data_at_time has been deprecated
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = ScalarData({m.var[:, "A"]: 5.5, m.input[:]: 6.6})
        interface.load_data(data)

        B_data = [m.var[t, "B"].value for t in m.time]
        # var[:,B] has not been changed
        self.assertEqual(B_data, [1.0, 1.1, 1.2])

        for t in m.time:
            self.assertEqual(m.var[t, "A"].value, 5.5)
            self.assertEqual(m.input[t].value, 6.6)

    def test_load_data_at_time_subset(self):
        # NOTE: load_data_at_time has been deprecated
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)

        old_A = {t: m.var[t, "A"].value for t in m.time}
        old_input = {t: m.input[t].value for t in m.time}

        data = ScalarData({m.var[:, "A"]: 5.5, m.input[:]: 6.6})
        time_points = [1, 2]
        time_set = set(time_points)
        interface.load_data(data, time_points=[1, 2])

        B_data = [m.var[t, "B"].value for t in m.time]
        # var[:,B] has not been changed
        self.assertEqual(B_data, [1.0, 1.1, 1.2])

        for t in m.time:
            if t in time_set:
                self.assertEqual(m.var[t, "A"].value, 5.5)
                self.assertEqual(m.input[t].value, 6.6)
            else:
                self.assertEqual(m.var[t, "A"].value, old_A[t])
                self.assertEqual(m.input[t].value, old_input[t])

    def test_load_data_from_dict_scalar_var(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = {pyo.ComponentUID(m.scalar): 6.0}
        interface.load_data(data)
        self.assertEqual(m.scalar.value, 6.0)

    def test_load_data_from_dict_indexed_var(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = {pyo.ComponentUID(m.input): 6.0}
        interface.load_data(data)
        for t in m.time:
            self.assertEqual(m.input[t].value, 6.0)

    def test_load_data_from_dict_indexed_var_list_data(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data_list = [2, 3, 4]
        data = {pyo.ComponentUID(m.input): data_list}
        # Need to provide data to load_data that can be interpreted
        # as a TimeSeriesData
        interface.load_data((data, m.time))
        for i, t in enumerate(m.time):
            self.assertEqual(m.input[t].value, data_list[i])

    def test_load_data_from_ScalarData_to_point(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = ScalarData({m.var[:, "A"]: 5.5, m.input[:]: 6.6})
        interface.load_data(data, time_points=1)

        B_data = [m.var[t, "B"].value for t in m.time]
        # var[:,B] has not been changed
        self.assertEqual(B_data, [1.0, 1.1, 1.2])

        old_A = [1.0, 1.1, 1.2]
        old_input = [1.0, 0.9, 0.8]
        for i, t in enumerate(m.time):
            if t == 1:
                self.assertEqual(m.var[t, "A"].value, 5.5)
                self.assertEqual(m.input[t].value, 6.6)
            else:
                self.assertEqual(m.var[t, "A"].value, old_A[i])
                self.assertEqual(m.input[t].value, old_input[i])

    def test_load_data_from_ScalarData_toall(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        data = ScalarData({m.var[:, "A"]: 5.5, m.input[:]: 6.6})
        interface.load_data(data)

        B_data = [m.var[t, "B"].value for t in m.time]
        # var[:,B] has not been changed
        self.assertEqual(B_data, [1.0, 1.1, 1.2])

        for t in m.time:
            self.assertEqual(m.var[t, "A"].value, 5.5)
            self.assertEqual(m.input[t].value, 6.6)

    def test_load_data_from_ScalarData_tosubset(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)

        old_A = {t: m.var[t, "A"].value for t in m.time}
        old_input = {t: m.input[t].value for t in m.time}

        data = ScalarData({m.var[:, "A"]: 5.5, m.input[:]: 6.6})
        time_points = [1, 2]
        time_set = set(time_points)
        interface.load_data(data, time_points=[1, 2])

        B_data = [m.var[t, "B"].value for t in m.time]
        # var[:,B] has not been changed
        self.assertEqual(B_data, [1.0, 1.1, 1.2])

        for t in m.time:
            if t in time_set:
                self.assertEqual(m.var[t, "A"].value, 5.5)
                self.assertEqual(m.input[t].value, 6.6)
            else:
                self.assertEqual(m.var[t, "A"].value, old_A[t])
                self.assertEqual(m.input[t].value, old_input[t])

    def test_load_data_from_TimeSeriesData(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        new_A = [1.0, 2.0, 3.0]
        new_input = [4.0, 5.0, 6.0]
        data = TimeSeriesData({m.var[:, "A"]: new_A, m.input[:]: new_input}, m.time)
        interface.load_data(data)

        B_data = [m.var[t, "B"].value for t in m.time]
        # var[:,B] has not been changed
        self.assertEqual(B_data, [1.0, 1.1, 1.2])

        for i, t in enumerate(m.time):
            self.assertEqual(m.var[t, "A"].value, new_A[i])
            self.assertEqual(m.input[t].value, new_input[i])

    def test_load_data_from_TimeSeriesData_tuple(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        new_A = [1.0, 2.0, 3.0]
        new_input = [4.0, 5.0, 6.0]
        data = ({m.var[:, "A"]: new_A, m.input[:]: new_input}, m.time)
        interface.load_data(data)

        B_data = [m.var[t, "B"].value for t in m.time]
        # var[:,B] has not been changed
        self.assertEqual(B_data, [1.0, 1.1, 1.2])

        for i, t in enumerate(m.time):
            self.assertEqual(m.var[t, "A"].value, new_A[i])
            self.assertEqual(m.input[t].value, new_input[i])

    def test_load_data_from_IntervalData(self):
        m = self._make_model(5)
        interface = DynamicModelInterface(m, m.time)
        new_A = [-1.1, -1.2, -1.3]
        new_input = [3.0, 2.9, 2.8]
        data = IntervalData(
            {m.var[:, "A"]: new_A, m.input[:]: new_input},
            [(0.0, 0.0), (0.0, 2.0), (2.0, 4.0)],
        )
        interface.load_data(data)
        B_data = [m.var[t, "B"].value for t in m.time]
        self.assertEqual(B_data, [1.0, 1.1, 1.2, 1.3, 1.4])
        for t in m.time:
            if t == 0:
                idx = 0
            elif t <= 2.0:
                idx = 1
            elif t <= 4.0:
                idx = 2
            self.assertEqual(m.var[t, "A"].value, new_A[idx])
            self.assertEqual(m.input[t].value, new_input[idx])

    def test_load_data_from_IntervalData_tuple(self):
        m = self._make_model(5)
        interface = DynamicModelInterface(m, m.time)
        new_A = [-1.1, -1.2, -1.3]
        new_input = [3.0, 2.9, 2.8]
        data = (
            {m.var[:, "A"]: new_A, m.input[:]: new_input},
            [(0.0, 0.0), (0.0, 2.0), (2.0, 4.0)],
        )
        interface.load_data(data)
        B_data = [m.var[t, "B"].value for t in m.time]
        self.assertEqual(B_data, [1.0, 1.1, 1.2, 1.3, 1.4])
        for t in m.time:
            if t == 0:
                idx = 0
            elif t <= 2.0:
                idx = 1
            elif t <= 4.0:
                idx = 2
            self.assertEqual(m.var[t, "A"].value, new_A[idx])
            self.assertEqual(m.input[t].value, new_input[idx])

    def test_load_data_bad_arg(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        new_A = [1.0, 2.0, 3.0]
        new_input = [4.0, 5.0, 6.0]
        data = ({m.var[:, "A"]: new_A, m.input[:]: new_input}, m.time)
        msg = "can only be set if data is IntervalData-compatible"
        with self.assertRaisesRegex(RuntimeError, msg):
            interface.load_data(data, prefer_left=True)

    def test_copy_values_at_time_default(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        interface.copy_values_at_time()
        # Default is to copy values from t0 to all points in time
        for t in m.time:
            self.assertEqual(m.var[t, "A"].value, 1.0)
            self.assertEqual(m.var[t, "B"].value, 1.0)
            self.assertEqual(m.input[t].value, 1.0)

    def test_copy_values_at_time_toall(self):
        m = self._make_model()
        tf = m.time.last()
        interface = DynamicModelInterface(m, m.time)
        interface.copy_values_at_time(source_time=tf)
        # Default is to copy values to all points in time
        for t in m.time:
            self.assertEqual(m.var[t, "A"].value, 1.2)
            self.assertEqual(m.var[t, "B"].value, 1.2)
            self.assertEqual(m.input[t].value, 0.8)

    def test_copy_values_at_time_tosubset(self):
        m = self._make_model()
        tf = m.time.last()
        interface = DynamicModelInterface(m, m.time)
        target_points = [t for t in m.time if t != m.time.first()]
        target_subset = set(target_points)
        interface.copy_values_at_time(source_time=tf, target_time=target_points)
        # Default is to copy values to all points in time
        for t in m.time:
            if t in target_subset:
                self.assertEqual(m.var[t, "A"].value, 1.2)
                self.assertEqual(m.var[t, "B"].value, 1.2)
                self.assertEqual(m.input[t].value, 0.8)
            else:
                # t0 has not been altered.
                self.assertEqual(m.var[t, "A"].value, 1.0)
                self.assertEqual(m.var[t, "B"].value, 1.0)
                self.assertEqual(m.input[t].value, 1.0)

    def test_copy_values_at_time_exception(self):
        m = self._make_model()
        tf = m.time.last()
        interface = DynamicModelInterface(m, m.time)
        msg = "copy_values_at_time can only copy"
        with self.assertRaisesRegex(ValueError, msg):
            interface.copy_values_at_time(source_time=m.time, target_time=tf)

    def test_shift_values_by_time(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        dt = 1.0
        interface.shift_values_by_time(dt)

        t = 0
        self.assertEqual(m.var[t, "A"].value, 1.1)
        self.assertEqual(m.var[t, "B"].value, 1.1)
        self.assertEqual(m.input[t].value, 0.9)

        t = 1
        self.assertEqual(m.var[t, "A"].value, 1.2)
        self.assertEqual(m.var[t, "B"].value, 1.2)
        self.assertEqual(m.input[t].value, 0.8)

        t = 2
        # For values within dt of the endpoint, the value at
        # the boundary is copied.
        self.assertEqual(m.var[t, "A"].value, 1.2)
        self.assertEqual(m.var[t, "B"].value, 1.2)
        self.assertEqual(m.input[t].value, 0.8)

    def test_get_penalty_from_constant_target(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        setpoint_data = ScalarData({m.var[:, "A"]: 1.0, m.var[:, "B"]: 2.0})
        weight_data = ScalarData({m.var[:, "A"]: 10.0, m.var[:, "B"]: 0.1})

        vset, tr_cost = interface.get_penalty_from_target(
            setpoint_data, weight_data=weight_data
        )
        m.var_set = vset
        m.tracking_cost = tr_cost
        for t in m.time:
            for i in m.var_set:
                pred_expr = (
                    10.0 * (m.var[t, "A"] - 1.0) ** 2
                    if i == 0
                    else 0.1 * (m.var[t, "B"] - 2.0) ** 2
                )
                self.assertEqual(pyo.value(pred_expr), pyo.value(m.tracking_cost[i, t]))
                self.assertTrue(
                    compare_expressions(pred_expr, m.tracking_cost[i, t].expr)
                )

    def test_get_penalty_from_constant_target_var_subset(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        setpoint_data = ScalarData(
            {m.var[:, "A"]: 1.0, m.var[:, "B"]: 2.0, m.input[:]: 3.0}
        )
        weight_data = ScalarData(
            {m.var[:, "A"]: 10.0, m.var[:, "B"]: 0.1, m.input[:]: 0.01}
        )

        variables = [m.var[:, "A"], m.var[:, "B"]]
        m.variable_set = pyo.Set(initialize=range(len(variables)))
        new_set, tr_cost = interface.get_penalty_from_target(
            setpoint_data,
            variables=variables,
            weight_data=weight_data,
            variable_set=m.variable_set,
        )
        m.tracking_cost = tr_cost
        self.assertIs(m.variable_set, new_set)
        for t in m.time:
            for i in m.variable_set:
                pred_expr = (
                    10.0 * (m.var[t, "A"] - 1.0) ** 2
                    if i == 0
                    else +0.1 * (m.var[t, "B"] - 2.0) ** 2
                )
                self.assertEqual(pyo.value(pred_expr), pyo.value(m.tracking_cost[i, t]))
                self.assertTrue(
                    compare_expressions(pred_expr, m.tracking_cost[i, t].expr)
                )


class TestGetPenaltyFromTarget(unittest.TestCase):
    def _make_model(self, n_time_points=3):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=list(range(n_time_points)))
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(
            m.time, m.comp, initialize={(i, j): 1.1 * i for i, j in m.time * m.comp}
        )
        return m

    def test_constant_setpoint(self):
        m = self._make_model()
        interface = DynamicModelInterface(m, m.time)
        setpoint = {m.var[:, "A"]: 0.3, m.var[:, "B"]: 0.4}
        m.var_set, m.penalty = interface.get_penalty_from_target(setpoint)

        # Note that the order of the variables here is not deterministic
        # in some Python <=3.6 implementations
        pred_expr = {
            (i, t): (
                (m.var[t, "A"] - 0.3) ** 2 if i == 0 else (m.var[t, "B"] - 0.4) ** 2
            )
            for i, t in m.var_set * m.time
        }
        for t in m.time:
            for i in m.var_set:
                self.assertTrue(
                    compare_expressions(pred_expr[i, t], m.penalty[i, t].expr)
                )
                self.assertEqual(pyo.value(pred_expr[i, t]), pyo.value(m.penalty[i, t]))

    def test_varying_setpoint(self):
        m = self._make_model(n_time_points=5)
        interface = DynamicModelInterface(m, m.time)
        A_target = [0.4, 0.6, 0.1, 0.0, 1.1]
        B_target = [0.8, 0.9, 1.3, 1.5, 1.4]
        setpoint = ({m.var[:, "A"]: A_target, m.var[:, "B"]: B_target}, m.time)
        m.var_set, m.penalty = interface.get_penalty_from_target(setpoint)

        target = {
            (i, t): A_target[j] if i == 0 else B_target[t]
            for i in m.var_set
            for (j, t) in enumerate(m.time)
        }
        for i, t in m.var_set * m.time:
            var = m.var[t, "A"] if i == 0 else m.var[t, "B"]
            pred_expr = (var - target[i, t]) ** 2
            self.assertTrue(compare_expressions(pred_expr, m.penalty[i, t].expr))
            self.assertEqual(pyo.value(pred_expr), pyo.value(m.penalty[i, t]))

    def test_piecewise_constant_setpoint(self):
        m = self._make_model(n_time_points=5)
        interface = DynamicModelInterface(m, m.time)
        A_target = [0.3, 0.9, 0.7]
        B_target = [1.1, 0.1, 0.5]
        setpoint = (
            {m.var[:, "A"]: A_target, m.var[:, "B"]: B_target},
            [(0.0, 0.0), (0.0, 2.0), (2.0, 4.0)],
        )
        m.var_set, m.penalty = interface.get_penalty_from_target(setpoint)
        target = {
            (i, j): A_target[j] if i == 0 else B_target[j]
            for i in m.var_set
            for j in range(len(A_target))
        }
        for i, t in m.var_set * m.time:
            if t == 0:
                idx = 0
            elif t <= 2.0:
                idx = 1
            elif t <= 4.0:
                idx = 2
            var = m.var[t, "A"] if i == 0 else m.var[t, "B"]
            pred_expr = (var - target[i, idx]) ** 2
            self.assertTrue(compare_expressions(pred_expr, m.penalty[i, t].expr))
            self.assertEqual(pyo.value(pred_expr), pyo.value(m.penalty[i, t]))

    def test_piecewise_constant_setpoint_with_specified_variables(self):
        m = self._make_model(n_time_points=5)
        interface = DynamicModelInterface(m, m.time)
        A_target = [0.3, 0.9, 0.7]
        B_target = [1.1, 0.1, 0.5]
        setpoint = (
            {m.var[:, "A"]: A_target, m.var[:, "B"]: B_target},
            [(0.0, 0.0), (0.0, 2.0), (2.0, 4.0)],
        )
        variables = [pyo.Reference(m.var[:, "B"])]
        m.var_set, m.penalty = interface.get_penalty_from_target(
            setpoint, variables=variables
        )
        self.assertEqual(len(m.var_set), 1)
        self.assertEqual(m.var_set[1], 0)
        for i, t in m.var_set * m.time:
            if t == 0:
                idx = 0
            elif t <= 2.0:
                idx = 1
            elif t <= 4.0:
                idx = 2
            var = m.var[t, "B"]
            pred_expr = (var - B_target[idx]) ** 2
            self.assertTrue(compare_expressions(pred_expr, m.penalty[i, t].expr))
            self.assertEqual(pyo.value(pred_expr), pyo.value(m.penalty[i, t]))

    def test_piecewise_constant_setpoint_time_subset(self):
        m = self._make_model(n_time_points=5)
        interface = DynamicModelInterface(m, m.time)
        A_target = [0.3, 0.9, 0.7]
        B_target = [1.1, 0.1, 0.5]
        setpoint = (
            {m.var[:, "A"]: A_target, m.var[:, "B"]: B_target},
            [(0.0, 0.0), (0.0, 2.0), (2.0, 4.0)],
        )
        m.sample_points = pyo.Set(initialize=[0.0, 2.0, 4.0])
        m.var_set, m.penalty = interface.get_penalty_from_target(
            setpoint, time=m.sample_points
        )
        idx_sets = ComponentSet(m.penalty.index_set().subsets())
        self.assertIn(m.var_set, idx_sets)
        self.assertIn(m.sample_points, idx_sets)
        target = {
            (i, j): A_target[j] if i == 0 else B_target[j]
            for i in m.var_set
            for j in range(len(A_target))
        }
        for i, t in m.var_set * m.time:
            if t == 0:
                idx = 0
            elif t <= 2.0:
                idx = 1
            elif t <= 4.0:
                idx = 2
            if t in m.sample_points:
                var = m.var[t, "A"] if i == 0 else m.var[t, "B"]
                pred_expr = (var - target[i, idx]) ** 2
                self.assertTrue(compare_expressions(pred_expr, m.penalty[i, t].expr))
                self.assertEqual(pyo.value(pred_expr), pyo.value(m.penalty[i, t]))
            else:
                self.assertNotIn((i, t), m.penalty)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


from typing import TYPE_CHECKING

from pyomo.common.factory import Factory
from pyomo.contrib.solver.common.base import LegacySolverWrapper
from pyomo.opt.base.solvers import LegacySolverFactory


class SolverFactoryClass(Factory):
    """Factory class for generating instances of solver interfaces (API v2)"""

    def register(self, name, legacy_name=None, doc=None):
        """Register a new solver with this solver factory

        This will register the solver both with this
        :attr:`SolverFactory` and with the original (legacy)
        :attr:`~pyomo.opt.base.solvers.LegacySolverFactory`

        Examples
        --------

        .. testcode::
           :hide:

           SolverFactory = SolverFactoryClass()

        This method can either be called as a decorator on a solver
        interface class definition, e.g.:

        .. testcode::

           from pyomo.contrib.solver.common.base import SolverBase

           @SolverFactory.register("test_solver_1")
           class TestSolver1(SolverBase):
               pass

        Or explicitly:

        .. testcode::

           class TestSolver2(SolverBase):
               pass

           SolverFactory.register("test_solver_2")(TestSolver2)

        When called explicitly, you can pass a custom class to register
        with the :attr:`LegacySolverFactory`:

        .. testcode::

           from pyomo.contrib.solver.common.base import LegacySolverWrapper

           class LegacyTestSolver2(LegacySolverWrapper, TestSolver2):
               pass

           SolverFactory.register("test_solver_2a")(TestSolver2, LegacyTestSolver2)


        Parameters
        ----------
        name : str

            The name used to register this solver interface class

        legacy_name : str

            The name to use to register the legacy interface wrapper to
            this solver interface in the LegacySolverInterface.  If
            ``None``, then ``name`` will be used.

        doc : str

            Extended description of this solver interface.

        """
        if legacy_name is None:
            legacy_name = name

        def decorator(cls, legacy_cls=None):
            self._cls[name] = cls
            self._doc[name] = doc

            if legacy_cls is None:

                class LegacySolver(LegacySolverWrapper, cls):
                    pass

                legacy_cls = LegacySolver

            LegacySolverFactory.register(
                legacy_name, doc + " (new interface)" if doc else doc
            )(legacy_cls)

            # Preserve the preferred name, as registered in the Factory
            cls.name = name
            return cls

        return decorator

    if TYPE_CHECKING:
        from pyomo.contrib.solver.common.base import SolverBase

        # NOTE: `Factory.__call__` can return None, but for the common case
        def __call__(self, name, **kwds) -> SolverBase: ...


#: Global registry/factory for "v2" solver interfaces.
SolverFactory: SolverFactoryClass = SolverFactoryClass()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Utility Functions
#

import pickle
import os
from os.path import abspath, dirname, join

currdir = dirname(abspath(__file__)) + os.sep
import platform

from filecmp import cmp
import pyomo.common.unittest as unittest
from pyomo.environ import (
    AbstractModel,
    ConcreteModel,
    Set,
    Param,
    Var,
    Constraint,
    Objective,
    Reals,
    NonNegativeReals,
    sum_product,
)

is_pypy = platform.python_implementation().lower().startswith("pypy")


def obj_rule(model):
    return sum(model.x[a] + model.y[a] for a in model.A)


def constr_rule(model, a):
    return model.x[a] >= model.y[a]


def simple_con_rule(model, i):
    return model.x <= i


class Test(unittest.TestCase):
    def verifyModel(self, ref, new):
        # Verify the block indices
        self.assertEqual(sorted(ref._data.keys()), sorted(new._data.keys()))
        for idx in ref._data.keys():
            self.assertEqual(type(ref._data[idx]), type(new._data[idx]))
            if idx is not None:
                self.assertNotEqual(id(ref._data[idx]), id(new._data[idx]))
        self.assertEqual(id(ref.solutions._instance()), id(ref))
        self.assertEqual(id(new.solutions._instance()), id(new))

        # Verify the block attributes
        for idx in ref._data.keys():
            # It is important to lookup using ref._data[idx] and not
            # ref[idx], as the model may not be constructed, and
            # accessing __getitem__ on an abstract model generates an
            # exception.
            ref_c = ref._data[idx].component_map()
            new_c = new._data[idx].component_map()
            self.assertEqual(sorted(ref_c.keys()), sorted(new_c.keys()))
            for a in ref_c.keys():
                self.assertEqual(type(ref_c[a]), type(new_c[a]))
                self.assertNotEqual(id(ref_c[a]), id(new_c[a]))

    def test_pickle_empty_abstract_model(self):
        model = AbstractModel()
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_set(self):
        model = AbstractModel()
        model.A = Set(initialize=[1, 2, 3])
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_virtual_set(self):
        model = AbstractModel()
        model._a = Set(initialize=[1, 2, 3])
        model.A = model._a * model._a
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_param(self):
        model = AbstractModel()
        model.A = Param(initialize=1)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_indexed_param(self):
        model = AbstractModel()
        model.A = Param([1, 2, 3], initialize={1: 100, 2: 200, 3: 300})
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_mutable_param(self):
        model = AbstractModel()
        model.A = Param(initialize=1, mutable=True)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_mutable_indexed_param(self):
        model = AbstractModel()
        model.A = Param([1, 2, 3], initialize={1: 100, 3: 300}, mutable=True)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_var(self):
        model = AbstractModel()
        model.A = Var(initialize=1)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_indexed_var(self):
        model = AbstractModel()
        model.A = Var([1, 2, 3], initialize={1: 100, 2: 200, 3: 300})
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_constant_objective(self):
        model = AbstractModel()
        model.A = Objective(expr=1)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_objective(self):
        model = AbstractModel()
        model.x = Var()
        model.A = Objective(expr=model.x <= 0)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_constraint(self):
        model = AbstractModel()
        model.x = Var()
        model.A = Constraint(expr=model.x <= 0)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_abstract_model_indexed_constraint(self):
        model = AbstractModel()
        model.x = Var()
        model.A = Constraint([1, 2, 3], rule=simple_con_rule)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    #########

    def test_pickle_empty_concrete_model(self):
        model = ConcreteModel()
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_set(self):
        model = ConcreteModel()
        model.A = Set(initialize=[1, 2, 3])
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_virtual_set(self):
        model = ConcreteModel()
        model._a = Set(initialize=[1, 2, 3])
        model.A = model._a * model._a
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_param(self):
        model = ConcreteModel()
        model.A = Param(initialize=1)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_indexed_param(self):
        model = ConcreteModel()
        model.A = Param([1, 2, 3], initialize={1: 100, 2: 200, 3: 300})
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_mutable_param(self):
        model = ConcreteModel()
        model.A = Param(initialize=1, mutable=True)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_mutable_indexed_param(self):
        model = ConcreteModel()
        model.A = Param([1, 2, 3], initialize={1: 100, 3: 300}, mutable=True)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_var(self):
        model = ConcreteModel()
        model.A = Var(initialize=1)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_indexed_var(self):
        model = ConcreteModel()
        model.A = Var([1, 2, 3], initialize={1: 100, 2: 200, 3: 300})
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_constant_objective(self):
        model = ConcreteModel()
        model.A = Objective(expr=1)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_objective(self):
        model = ConcreteModel()
        model.x = Var()
        model.A = Objective(expr=model.x**2)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_constraint(self):
        model = ConcreteModel()
        model.x = Var()
        model.A = Constraint(expr=model.x <= 0)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    def test_pickle_concrete_model_indexed_constraint(self):
        model = ConcreteModel()
        model.x = Var()
        model.A = Constraint([1, 2, 3], rule=simple_con_rule)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        self.verifyModel(model, tmodel)

    ##########

    # tests the ability to pickle an abstract model prior to construction,
    # read it back it, and create an instance from it. validation is relatively
    # weak, in that it only tests the validity of an expression constructed
    # using the resulting model.
    def test_pickle1(self):
        model = AbstractModel()
        model.A = Set(initialize=[1, 2, 3])
        model.B = Param(model.A, initialize={1: 100, 2: 200, 3: 300}, mutable=True)
        model.x = Var(model.A)
        model.y = Var(model.A)
        model.obj = Objective(rule=obj_rule)
        model.constr = Constraint(model.A, rule=constr_rule)
        pickle_str = pickle.dumps(model)
        tmodel = pickle.loads(pickle_str)
        instance = tmodel.create_instance()
        expr = sum_product(instance.x, instance.B, instance.y)
        baseline = "B[1]*x[1]*y[1] + B[2]*x[2]*y[2] + B[3]*x[3]*y[3]"
        self.assertEqual(str(expr), baseline)

    # same as above, but pickles the constructed AbstractModel and
    # then operates on the unpickled instance.
    def test_pickle2(self):
        model = AbstractModel()
        model.A = Set(initialize=[1, 2, 3])
        model.B = Param(model.A, initialize={1: 100, 2: 200, 3: 300}, mutable=True)
        model.x = Var(model.A)
        model.y = Var(model.A)
        model.obj = Objective(rule=obj_rule)
        model.constr = Constraint(model.A, rule=constr_rule)
        tmp = model.create_instance()
        pickle_str = pickle.dumps(tmp)
        instance = pickle.loads(pickle_str)
        expr = sum_product(instance.x, instance.B, instance.y)
        baseline = "B[1]*x[1]*y[1] + B[2]*x[2]*y[2] + B[3]*x[3]*y[3]"
        self.assertEqual(str(expr), baseline)

    # verifies that the use of lambda expressions as rules yields model instances
    # that are not pickle'able.
    def test_pickle3(self):
        def rule1(model):
            return (1, model.x + model.y[1], 2)

        def rule2(model, i):
            return (1, model.x + model.y[1] + i, 2)

        model = AbstractModel()
        model.a = Set(initialize=[1, 2, 3])
        model.A = Param(initialize=1, mutable=True)
        model.B = Param(model.a, mutable=True)
        model.x = Var(initialize=1, within=Reals)
        model.y = Var(model.a, initialize=1, within=Reals)
        model.obj = Objective(rule=lambda model: model.x + model.y[1])
        model.obj2 = Objective(model.a, rule=lambda model, i: i + model.x + model.y[1])
        model.con = Constraint(rule=rule1)
        model.con2 = Constraint(model.a, rule=rule2)
        instance = model.create_instance()
        if is_pypy:
            str_ = pickle.dumps(instance)
            tmp_ = pickle.loads(str_)
        else:
            with self.assertRaises((pickle.PicklingError, TypeError, AttributeError)):
                pickle.dumps(instance)

    # verifies that we can print a constructed model and
    # obtain identical results before and after
    # pickling. introduced due to a test case by Gabe that
    # illustrated __getstate__ of various modeling
    # components was incorrectly and unexpectedly modifying
    # object state.
    def test_pickle4(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1, 2])
        model.x = Var(within=NonNegativeReals)
        model.x_indexed = Var(model.s, within=NonNegativeReals)
        model.obj = Objective(expr=model.x + model.x_indexed[1] + model.x_indexed[2])
        model.con = Constraint(expr=model.x >= 1)
        model.con2 = Constraint(expr=model.x_indexed[1] + model.x_indexed[2] >= 4)

        OUTPUT = open(join(currdir, "test_pickle4_baseline.out"), "w")
        model.pprint(ostream=OUTPUT)
        OUTPUT.close()
        _out, _txt = join(currdir, "test_pickle4_baseline.out"), join(
            currdir, "test_pickle4_baseline.txt"
        )
        self.assertTrue(cmp(_out, _txt), msg="Files %s and %s differ" % (_out, _txt))

        str = pickle.dumps(model)

        OUTPUT = open(join(currdir, "test_pickle4_after.out"), "w")
        model.pprint(ostream=OUTPUT)
        OUTPUT.close()
        _out, _txt = join(currdir, "test_pickle4_after.out"), join(
            currdir, "test_pickle4_baseline.txt"
        )
        self.assertTrue(cmp(_out, _txt), msg="Files %s and %s differ" % (_out, _txt))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import pyomo.common.unittest as unittest

from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy_available,
)

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run BlockVector tests")

from pyomo.contrib.pynumero.sparse.block_vector import (
    BlockVector,
    NotFullyDefinedBlockVectorError,
    vec_associative_reductions,
    vec_unary_ufuncs,
    vec_binary_ufuncs,
)


class TestBlockVector(unittest.TestCase):
    def test_constructor(self):
        v = BlockVector(2)
        self.assertEqual(v.nblocks, 2)
        self.assertEqual(v.bshape, (2,))
        with self.assertRaises(NotFullyDefinedBlockVectorError):
            v_size = v.size

        v.set_block(0, np.ones(2))
        v.set_block(1, np.ones(4))
        self.assertEqual(v.size, 6)
        self.assertEqual(v.shape, (6,))
        with self.assertRaises(AssertionError):
            v.set_block(0, None)

        with self.assertRaises(Exception) as context:
            BlockVector('hola')

    def setUp(self):
        self.ones = BlockVector(3)
        self.list_sizes_ones = [2, 4, 3]
        for idx, s in enumerate(self.list_sizes_ones):
            self.ones.set_block(idx, np.ones(s))

    def test_block_sizes(self):
        self.assertListEqual(self.ones.block_sizes().tolist(), self.list_sizes_ones)

    def test_dot(self):
        v1 = self.ones
        self.assertEqual(v1.dot(v1), v1.size)
        v2 = v1.clone(3.3)
        self.assertAlmostEqual(v1.dot(v2), v1.size * 3.3)
        self.assertAlmostEqual(v2.dot(v1.flatten()), v1.size * 3.3)
        with self.assertRaises(Exception) as context:
            v1.dot(1.0)

    def test_mean(self):
        flat_v = np.ones(self.ones.size)
        v = self.ones
        self.assertEqual(v.mean(), flat_v.mean())
        v = BlockVector(2)
        with self.assertRaises(NotFullyDefinedBlockVectorError):
            v_mean = v.mean()

    def test_sum(self):
        self.assertEqual(self.ones.sum(), self.ones.size)
        v = BlockVector(2)
        v.set_block(0, np.arange(5))
        v.set_block(1, np.arange(9))
        self.assertEqual(v.sum(), 46)

    def test_all(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.ones(3)
        v.set_block(0, a)
        v.set_block(1, b)
        self.assertTrue(v.all())

        v = BlockVector(2)
        a = np.zeros(5)
        b = np.zeros(3)
        v.set_block(0, a)
        v.set_block(1, b)
        self.assertFalse(v.all())

    def test_any(self):
        v = BlockVector(2)
        a = np.zeros(5)
        b = np.ones(3)
        v.set_block(0, a)
        v.set_block(1, b)
        self.assertTrue(v.any())

        v = BlockVector(2)
        a = np.zeros(5)
        b = np.zeros(3)
        v.set_block(0, a)
        v.set_block(1, b)
        self.assertFalse(v.any())

    def test_argpartition(self):
        v = self.ones
        self.assertRaises(NotImplementedError, v.argpartition, 1)

    def test_argsort(self):
        v = self.ones
        self.assertRaises(NotImplementedError, v.argsort)

    def test_astype(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.astype(np.int64, copy=False)

        vv = v.astype(np.int64)
        self.assertEqual(vv.nblocks, v.nblocks)
        for blk in vv:
            self.assertEqual(blk.dtype, np.int64)

    def test_byteswap(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            v.byteswap()

    def test_choose(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            v.choose(1)

    def test_clip(self):
        v = BlockVector(3)
        v2 = BlockVector(3)
        a = np.zeros(5)
        b = np.ones(3) * 5.0
        c = np.ones(3) * 10.0

        v.set_block(0, a)
        v.set_block(1, b)
        v.set_block(2, c)

        v2.set_block(0, np.ones(5) * 4.0)
        v2.set_block(1, np.ones(3) * 5.0)
        v2.set_block(2, np.ones(3) * 9.0)

        vv = v.clip(4.0, 9.0)
        self.assertEqual(vv.nblocks, v.nblocks)
        for bid, blk in enumerate(vv):
            self.assertTrue(np.allclose(blk, v2.get_block(bid)))

    def test_compress(self):
        v = self.ones

        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)

        c = v.compress(v < 1)

        v2 = BlockVector(2)
        b = np.zeros(9)
        v2.set_block(0, np.ones(0))
        v2.set_block(1, b)

        self.assertEqual(c.nblocks, v.nblocks)
        for bid, blk in enumerate(c):
            self.assertTrue(np.allclose(blk, v2.get_block(bid)))

        flags = v < 1
        c = v.compress(flags.flatten())
        self.assertEqual(c.nblocks, v.nblocks)
        for bid, blk in enumerate(c):
            self.assertTrue(np.allclose(blk, v2.get_block(bid)))

        with self.assertRaises(Exception) as context:
            v.compress(1.0)

    def test_nonzero(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)

        n = v.nonzero()

        v2 = BlockVector(2)
        v2.set_block(0, np.arange(5))
        v2.set_block(1, np.zeros(0))
        self.assertEqual(n[0].nblocks, v.nblocks)
        for bid, blk in enumerate(n[0]):
            self.assertTrue(np.allclose(blk, v2.get_block(bid)))

    @unittest.skipUnless(np.__version__[0] == "1", "PTP only included in Numpy 1.x")
    def test_ptp(self):
        v = BlockVector(2)
        a = np.arange(5)
        b = np.arange(9)
        v.set_block(0, a)
        v.set_block(1, b)

        vv = np.arange(9)
        self.assertEqual(vv.ptp(), v.ptp())

    def test_round(self):
        v = BlockVector(2)
        a = np.ones(5) * 1.1
        b = np.ones(9) * 1.1
        v.set_block(0, a)
        v.set_block(1, b)

        vv = v.round()
        self.assertEqual(vv.nblocks, v.nblocks)
        a = np.ones(5)
        b = np.ones(9)
        v.set_block(0, a)
        v.set_block(1, b)
        for bid, blk in enumerate(vv):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

    def test_std(self):
        v = BlockVector(2)
        a = np.arange(5)
        b = np.arange(9)
        v.set_block(0, a)
        v.set_block(1, b)

        vv = np.concatenate([a, b])
        self.assertEqual(vv.std(), v.std())

    def test_conj(self):
        v = self.ones
        vv = v.conj()
        self.assertEqual(vv.nblocks, v.nblocks)
        self.assertEqual(vv.shape, v.shape)
        for bid, blk in enumerate(vv):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

    def test_conjugate(self):
        v = self.ones
        vv = v.conjugate()
        self.assertEqual(vv.nblocks, v.nblocks)
        self.assertEqual(vv.shape, v.shape)
        for bid, blk in enumerate(vv):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

    def test_diagonal(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.diagonal()

    def test_getfield(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.getfield(1)

    def test_item(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.item(1)

    def test_itemset(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.itemset(1)

    def test_newbyteorder(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.newbyteorder()

    def test_partition(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.partition(1)

    def test_repeat(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.repeat(1)

    def test_reshape(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.reshape(1)

    def test_resize(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.resize(1)

    def test_searchsorted(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.searchsorted(1)

    def test_setfield(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.setfield(1, 1)

    def test_setflags(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.setflags()

    def test_sort(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.sort()

    def test_squeeze(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.squeeze()

    def test_swapaxes(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.swapaxes(1, 1)

    def test_tobytes(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.tobytes()

    def test_trace(self):
        v = self.ones
        with self.assertRaises(NotImplementedError) as ctx:
            vv = v.trace()

    def test_prod(self):
        self.assertEqual(self.ones.prod(), 1)
        v = BlockVector(2)
        a = np.arange(5)
        b = np.arange(9)
        c = np.concatenate([a, b])
        v.set_block(0, a)
        v.set_block(1, b)
        self.assertEqual(v.prod(), c.prod())

    def test_max(self):
        self.assertEqual(self.ones.max(), 1)
        v = BlockVector(2)
        a = np.arange(5)
        b = np.arange(9)
        c = np.concatenate([a, b])
        v.set_block(0, a)
        v.set_block(1, b)
        self.assertEqual(v.max(), c.max())

    def test_min(self):
        self.assertEqual(self.ones.min(), 1)
        v = BlockVector(2)
        a = np.arange(5)
        b = np.arange(9)
        c = np.concatenate([a, b])
        v.set_block(0, a)
        v.set_block(1, b)
        self.assertEqual(v.min(), c.min())

    def test_tolist(self):
        v = BlockVector(2)
        a = np.arange(5)
        b = np.arange(9)
        c = np.concatenate([a, b])
        v.set_block(0, a)
        v.set_block(1, b)
        self.assertListEqual(v.tolist(), c.tolist())

    def test_flatten(self):
        v = BlockVector(2)
        a = np.arange(5)
        b = np.arange(9)
        c = np.concatenate([a, b])
        v.set_block(0, a)
        v.set_block(1, b)
        self.assertListEqual(v.flatten().tolist(), c.tolist())

    def test_fill(self):
        v = BlockVector(2)
        a = np.arange(5)
        b = np.arange(9)
        v.set_block(0, a)
        v.set_block(1, b)
        v.fill(1.0)
        c = np.ones(v.size)
        self.assertListEqual(v.tolist(), c.tolist())

    def test_shape(self):
        size = sum(self.list_sizes_ones)
        self.assertEqual(self.ones.shape, (size,))

    def test_bshape(self):
        self.assertEqual(self.ones.bshape, (3,))

    def test_size(self):
        size = sum(self.list_sizes_ones)
        self.assertEqual(self.ones.size, size)

    def test_length(self):
        size = sum(self.list_sizes_ones)
        self.assertEqual(len(self.ones), self.ones.nblocks)

    def test_argmax(self):
        v = BlockVector(3)
        a = np.array([3, 2, 1])
        v.set_block(0, a.copy())
        v.set_block(1, a.copy())
        v.set_block(2, a.copy())
        v.get_block(1)[1] = 5
        argmax = v.argmax()
        self.assertEqual(argmax, 4)

    def test_argmin(self):
        v = BlockVector(3)
        a = np.array([3, 2, 1])
        v.set_block(0, a.copy())
        v.set_block(1, a.copy())
        v.set_block(2, a.copy())
        v.get_block(1)[1] = -5
        argmin = v.argmin()
        self.assertEqual(argmin, 4)

    def test_cumprod(self):
        v = BlockVector(3)
        v.set_block(0, np.arange(1, 5))
        v.set_block(1, np.arange(5, 10))
        v.set_block(2, np.arange(10, 15))
        c = np.arange(1, 15)
        res = v.cumprod()
        self.assertIsInstance(res, BlockVector)
        self.assertEqual(v.nblocks, res.nblocks)
        self.assertTrue(np.allclose(c.cumprod(), res.flatten()))

    def test_cumsum(self):
        v = BlockVector(3)
        v.set_block(0, np.arange(1, 5))
        v.set_block(1, np.arange(5, 10))
        v.set_block(2, np.arange(10, 15))
        c = np.arange(1, 15)
        res = v.cumsum()
        self.assertIsInstance(res, BlockVector)
        self.assertEqual(v.nblocks, res.nblocks)
        self.assertTrue(np.allclose(c.cumsum(), res.flatten()))

    def test_clone(self):
        v = self.ones
        w = v.clone()
        self.assertListEqual(w.tolist(), v.tolist())
        x = v.clone(4)
        self.assertListEqual(x.tolist(), [4] * v.size)
        y = x.clone(copy=False)
        y.get_block(2)[-1] = 6
        d = np.ones(y.size) * 4
        d[-1] = 6
        self.assertListEqual(y.tolist(), d.tolist())
        self.assertListEqual(x.tolist(), d.tolist())

    def test_add(self):
        v = self.ones
        v1 = self.ones
        result = v + v1
        self.assertListEqual(result.tolist(), [2] * v.size)
        result = v + 2
        self.assertListEqual(result.tolist(), [3] * v.size)
        result = v + v1.flatten()
        self.assertTrue(np.allclose(result.flatten(), v.flatten() + v1.flatten()))

        with self.assertRaises(Exception) as context:
            result = v + 'hola'

    def test_radd(self):
        v = self.ones
        v1 = self.ones
        result = v + v1
        self.assertListEqual(result.tolist(), [2] * v.size)
        result = 2 + v
        self.assertListEqual(result.tolist(), [3] * v.size)
        result = v.flatten() + v1
        self.assertTrue(np.allclose(result.flatten(), v.flatten() + v1.flatten()))

    def test_sub(self):
        v = self.ones
        v1 = self.ones
        result = v - v1
        self.assertListEqual(result.tolist(), [0] * v.size)
        result = v - 1.0
        self.assertListEqual(result.tolist(), [0] * v.size)
        result = v - v1.flatten()
        self.assertTrue(np.allclose(result.flatten(), v.flatten() - v1.flatten()))

        with self.assertRaises(Exception) as context:
            result = v - 'hola'

    def test_rsub(self):
        v = self.ones
        v1 = self.ones
        result = v1.__rsub__(v)
        self.assertListEqual(result.tolist(), [0] * v.size)
        result = 1.0 - v
        self.assertListEqual(result.tolist(), [0] * v.size)
        result = v.flatten() - v1
        self.assertTrue(np.allclose(result.flatten(), v.flatten() - v1.flatten()))

    def test_mul(self):
        v = self.ones
        v1 = v.clone(5, copy=True)
        result = v1 * v
        self.assertListEqual(result.tolist(), [5] * v.size)
        result = v * 5.0
        self.assertListEqual(result.tolist(), [5] * v.size)
        result = v * v1.flatten()
        self.assertTrue(np.allclose(result.flatten(), v.flatten() * v1.flatten()))

        with self.assertRaises(Exception) as context:
            result = v * 'hola'

    def test_rmul(self):
        v = self.ones
        v1 = v.clone(5, copy=True)
        result = v1.__rmul__(v)
        self.assertListEqual(result.tolist(), [5] * v.size)
        result = v * 5.0
        self.assertListEqual(result.tolist(), [5] * v.size)
        result = v.flatten() * v1
        self.assertTrue(np.allclose(result.flatten(), v.flatten() * v1.flatten()))

    def test_truediv(self):
        v = self.ones
        v1 = v.clone(5.0, copy=True)
        result = v / v1
        self.assertListEqual(result.tolist(), [1.0 / 5.0] * v.size)
        result = v / v1.flatten()
        self.assertTrue(np.allclose(result.flatten(), v.flatten() / v1.flatten()))
        result = 5.0 / v1
        self.assertTrue(np.allclose(result.flatten(), v.flatten()))
        result = v1 / 5.0
        self.assertTrue(np.allclose(result.flatten(), v.flatten()))

    def test_rtruediv(self):
        v = self.ones
        v1 = v.clone(5.0, copy=True)
        result = v1.__rtruediv__(v)
        self.assertListEqual(result.tolist(), [1.0 / 5.0] * v.size)
        result = v.flatten() / v1
        self.assertTrue(np.allclose(result.flatten(), v.flatten() / v1.flatten()))
        result = 5.0 / v1
        self.assertTrue(np.allclose(result.flatten(), v.flatten()))
        result = v1 / 5.0
        self.assertTrue(np.allclose(result.flatten(), v.flatten()))

    def test_floordiv(self):
        v = self.ones
        v.fill(2.0)
        v1 = v.clone(5.0, copy=True)
        result = v1 // v
        self.assertListEqual(result.tolist(), [5.0 // 2.0] * v.size)
        result = v // v1.flatten()
        self.assertTrue(np.allclose(result.flatten(), v.flatten() // v1.flatten()))

    def test_rfloordiv(self):
        v = self.ones
        v.fill(2.0)
        v1 = v.clone(5.0, copy=True)
        result = v.__rfloordiv__(v1)
        self.assertListEqual(result.tolist(), [5.0 // 2.0] * v.size)
        result = v.flatten() // v1
        self.assertTrue(np.allclose(result.flatten(), v.flatten() // v1.flatten()))
        result = 2.0 // v1
        self.assertTrue(np.allclose(result.flatten(), np.zeros(v1.size)))
        result = v1 // 2.0
        self.assertTrue(np.allclose(result.flatten(), np.ones(v1.size) * 2.0))

    def test_iadd(self):
        v = self.ones
        v += 3
        self.assertListEqual(v.tolist(), [4] * v.size)
        v.fill(1.0)
        v += v
        self.assertListEqual(v.tolist(), [2] * v.size)
        v.fill(1.0)
        v += np.ones(v.size) * 3
        self.assertTrue(np.allclose(v.flatten(), np.ones(v.size) * 4))

        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        a_copy = a.copy()
        b_copy = b.copy()

        v.set_block(0, a)
        v.set_block(1, b)
        v += 1.0

        self.assertTrue(np.allclose(v.get_block(0), a_copy + 1))
        self.assertTrue(np.allclose(v.get_block(1), b_copy + 1))

        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        a_copy = a.copy()
        b_copy = b.copy()

        v.set_block(0, a)
        v.set_block(1, b)

        v2 = BlockVector(2)
        v2.set_block(0, np.ones(5))
        v2.set_block(1, np.ones(9))

        v += v2
        self.assertTrue(np.allclose(v.get_block(0), a_copy + 1))
        self.assertTrue(np.allclose(v.get_block(1), b_copy + 1))

        self.assertTrue(np.allclose(v2.get_block(0), np.ones(5)))
        self.assertTrue(np.allclose(v2.get_block(1), np.ones(9)))

        with self.assertRaises(Exception) as context:
            v += 'hola'

    def test_isub(self):
        v = self.ones
        v -= 3
        self.assertListEqual(v.tolist(), [-2] * v.size)
        v.fill(1.0)
        v -= v
        self.assertListEqual(v.tolist(), [0] * v.size)
        v.fill(1.0)
        v -= np.ones(v.size) * 3
        self.assertTrue(np.allclose(v.flatten(), -np.ones(v.size) * 2))

        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        a_copy = a.copy()
        b_copy = b.copy()
        v.set_block(0, a)
        v.set_block(1, b)
        v -= 5.0

        self.assertTrue(np.allclose(v.get_block(0), a_copy - 5.0))
        self.assertTrue(np.allclose(v.get_block(1), b_copy - 5.0))

        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        a_copy = a.copy()
        b_copy = b.copy()
        v.set_block(0, a)
        v.set_block(1, b)

        v2 = BlockVector(2)
        v2.set_block(0, np.ones(5))
        v2.set_block(1, np.ones(9))

        v -= v2
        self.assertTrue(np.allclose(v.get_block(0), a_copy - 1))
        self.assertTrue(np.allclose(v.get_block(1), b_copy - 1))

        self.assertTrue(np.allclose(v2.get_block(0), np.ones(5)))
        self.assertTrue(np.allclose(v2.get_block(1), np.ones(9)))

        with self.assertRaises(Exception) as context:
            v -= 'hola'

    def test_imul(self):
        v = self.ones
        v *= 3
        self.assertListEqual(v.tolist(), [3] * v.size)
        v.fill(1.0)
        v *= v
        self.assertListEqual(v.tolist(), [1] * v.size)
        v.fill(1.0)
        v *= np.ones(v.size) * 2
        self.assertTrue(np.allclose(v.flatten(), np.ones(v.size) * 2))

        v = BlockVector(2)
        a = np.ones(5)
        b = np.arange(9, dtype=np.float64)
        a_copy = a.copy()
        b_copy = b.copy()
        v.set_block(0, a)
        v.set_block(1, b)
        v *= 2.0

        self.assertTrue(np.allclose(v.get_block(0), a_copy * 2.0))
        self.assertTrue(np.allclose(v.get_block(1), b_copy * 2.0))

        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        a_copy = a.copy()
        b_copy = b.copy()
        v.set_block(0, a)
        v.set_block(1, b)

        v2 = BlockVector(2)
        v2.set_block(0, np.ones(5) * 2)
        v2.set_block(1, np.ones(9) * 2)

        v *= v2
        self.assertTrue(np.allclose(v.get_block(0), a_copy * 2))
        self.assertTrue(np.allclose(v.get_block(1), b_copy * 2))

        self.assertTrue(np.allclose(v2.get_block(0), np.ones(5) * 2))
        self.assertTrue(np.allclose(v2.get_block(1), np.ones(9) * 2))

        with self.assertRaises(Exception) as context:
            v *= 'hola'

    def test_itruediv(self):
        v = self.ones
        v /= 3
        self.assertTrue(np.allclose(v.flatten(), np.ones(v.size) / 3))
        v.fill(1.0)
        v /= v
        self.assertTrue(np.allclose(v.flatten(), np.ones(v.size)))
        v.fill(1.0)
        v /= np.ones(v.size) * 2
        self.assertTrue(np.allclose(v.flatten(), np.ones(v.size) / 2))

        v = BlockVector(2)
        a = np.ones(5)
        b = np.arange(9, dtype=np.float64)
        a_copy = a.copy()
        b_copy = b.copy()
        v.set_block(0, a)
        v.set_block(1, b)
        v /= 2.0

        self.assertTrue(np.allclose(v.get_block(0), a_copy / 2.0))
        self.assertTrue(np.allclose(v.get_block(1), b_copy / 2.0))

        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        a_copy = a.copy()
        b_copy = b.copy()
        v.set_block(0, a)
        v.set_block(1, b)

        v2 = BlockVector(2)
        v2.set_block(0, np.ones(5) * 2)
        v2.set_block(1, np.ones(9) * 2)

        v /= v2
        self.assertTrue(np.allclose(v.get_block(0), a_copy / 2))
        self.assertTrue(np.allclose(v.get_block(1), b_copy / 2))

        self.assertTrue(np.allclose(v2.get_block(0), np.ones(5) * 2))
        self.assertTrue(np.allclose(v2.get_block(1), np.ones(9) * 2))

        with self.assertRaises(Exception) as context:
            v *= 'hola'

    def test_getitem(self):
        v = self.ones
        for i, s in enumerate(self.list_sizes_ones):
            self.assertEqual(v.get_block(i).size, s)
            self.assertEqual(v.get_block(i).shape, (s,))
            self.assertListEqual(v.get_block(i).tolist(), np.ones(s).tolist())

    def test_setitem(self):
        v = self.ones
        for i, s in enumerate(self.list_sizes_ones):
            v.set_block(i, np.ones(s) * i)
        for i, s in enumerate(self.list_sizes_ones):
            self.assertEqual(v.get_block(i).size, s)
            self.assertEqual(v.get_block(i).shape, (s,))
            res = np.ones(s) * i
            self.assertListEqual(v.get_block(i).tolist(), res.tolist())

    def test_set_blocks(self):
        v = self.ones
        blocks = [np.ones(s) * i for i, s in enumerate(self.list_sizes_ones)]
        v.set_blocks(blocks)
        for i, s in enumerate(self.list_sizes_ones):
            self.assertEqual(v.get_block(i).size, s)
            self.assertEqual(v.get_block(i).shape, (s,))
            res = np.ones(s) * i
            self.assertListEqual(v.get_block(i).tolist(), res.tolist())

    def test_has_none(self):
        v = self.ones
        self.assertFalse(v.has_none)
        v = BlockVector(3)
        v.set_block(0, np.ones(2))
        v.set_block(2, np.ones(3))
        self.assertTrue(v.has_none)
        v.set_block(1, np.ones(2))
        self.assertFalse(v.has_none)

    def test_copyfrom(self):
        v = self.ones
        v1 = np.zeros(v.size)
        v.copyfrom(v1)
        self.assertListEqual(v.tolist(), v1.tolist())

        v2 = BlockVector(len(self.list_sizes_ones))
        for i, s in enumerate(self.list_sizes_ones):
            v2.set_block(i, np.ones(s) * i)
        v.copyfrom(v2)
        for idx, blk in enumerate(v2):
            self.assertListEqual(blk.tolist(), v2.get_block(idx).tolist())

        v3 = BlockVector(2)
        v4 = v.clone(2)
        v3.set_block(0, v4)
        v3.set_block(1, np.zeros(3))
        self.assertListEqual(v3.tolist(), v4.tolist() + [0] * 3)

    def test_copyto(self):
        v = self.ones
        v2 = BlockVector(len(self.list_sizes_ones))
        v.copyto(v2)
        self.assertListEqual(v.tolist(), v2.tolist())
        v3 = np.zeros(v.size)
        v.copyto(v3)
        self.assertListEqual(v.tolist(), v3.tolist())
        v *= 5
        v.copyto(v2)
        self.assertListEqual(v.tolist(), v2.tolist())

    def test_gt(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)

        flags = v > 0
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

        flags = v > np.zeros(v.size)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

        vv = v.copy()
        vv.fill(0.0)
        flags = v > vv
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

    def test_ge(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)

        flags = v >= 0
        v.set_block(1, b + 1)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))
        v.set_block(1, b - 1)
        flags = v >= np.zeros(v.size)
        v.set_block(1, b)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

        v.set_block(1, b - 1)
        vv = v.copy()
        vv.fill(0.0)
        flags = v >= vv
        v.set_block(1, b)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

    def test_lt(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)

        flags = v < 1
        v.set_block(0, a - 1)
        v.set_block(1, b + 1)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))
        v.set_block(0, a + 1)
        v.set_block(1, b - 1)
        flags = v < np.ones(v.size)
        v.set_block(0, a - 1)
        v.set_block(1, b + 1)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

        v.set_block(0, a + 1)
        v.set_block(1, b - 1)
        vv = v.copy()
        vv.fill(1.0)
        flags = v < vv
        v.set_block(0, a - 1)
        v.set_block(1, b + 1)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

    def test_le(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)

        flags = v <= 1
        v.set_block(1, b + 1)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

        flags = v <= v
        vv = v.copy()
        vv.fill(1.0)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, vv.get_block(bid)))

        flags = v <= v.flatten()
        vv = v.copy()
        vv.fill(1.0)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, vv.get_block(bid)))

    def test_eq(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)

        flags = v == 1
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

        flags = v == np.ones(v.size)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

        vv = v.copy()
        vv.fill(1.0)
        flags = v == vv
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

    def test_ne(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)

        flags = v != 0
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

        flags = v != np.zeros(v.size)
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

        vv = v.copy()
        vv.fill(0.0)
        flags = v != vv
        self.assertEqual(v.nblocks, flags.nblocks)
        for bid, blk in enumerate(flags):
            self.assertTrue(np.allclose(blk, v.get_block(bid)))

    def test_contains(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)

        self.assertTrue(0 in v)
        self.assertFalse(3 in v)

    # ToDo: add tests for block vectors with block vectors in them
    # ToDo: vector comparisons
    def test_copy(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)
        v2 = v.copy()
        self.assertTrue(np.allclose(v.flatten(), v2.flatten()))

    def test_copy_structure(self):
        v = BlockVector(2)
        a = np.ones(5)
        b = np.zeros(9)
        v.set_block(0, a)
        v.set_block(1, b)
        v2 = v.copy_structure()
        self.assertEqual(v.get_block(0).size, v2.get_block(0).size)
        self.assertEqual(v.get_block(1).size, v2.get_block(1).size)

    def test_unary_ufuncs(self):
        a = np.ones(3) * 0.5
        b = np.ones(2) * 0.8
        v = BlockVector(2)
        v.set_block(0, a)
        v.set_block(1, b)
        # Some operations only accept integers
        ai = np.ones(3, dtype='i') * 5
        bi = np.ones(2, dtype='i') * 8
        vi = BlockVector(2)
        vi.set_block(0, ai)
        vi.set_block(1, bi)

        _int_ufuncs = {np.invert, np.arccosh}

        v2 = BlockVector(2)
        for fun in vec_unary_ufuncs:
            _v = vi if fun in _int_ufuncs else v
            v2.set_block(0, fun(_v.get_block(0)))
            v2.set_block(1, fun(_v.get_block(1)))
            res = fun(_v)
            self.assertIsInstance(res, BlockVector)
            self.assertEqual(res.nblocks, 2)
            for i in range(2):
                self.assertTrue(np.allclose(res.get_block(i), v2.get_block(i)))

        other_funcs = [np.cumsum, np.cumprod]
        if np.__version__[0] == '1':
            other_funcs.append(np.cumproduct)

        for fun in other_funcs:
            res = fun(v)
            self.assertIsInstance(res, BlockVector)
            self.assertEqual(res.nblocks, 2)
            self.assertTrue(np.allclose(fun(v.flatten()), res.flatten()))

        with self.assertRaises(TypeError):
            np.modf(v)

    def test_reduce_ufuncs(self):
        v = BlockVector(2)
        # Some operations only accept integers, so we will test with integers
        a = np.ones(3, dtype='i') * 5
        b = np.ones(2, dtype='i') * 8
        v.set_block(0, a)
        v.set_block(1, b)

        for fun in vec_associative_reductions:
            self.assertAlmostEqual(fun.reduce(v), fun.reduce(v.flatten()))

        reduce_funcs = [np.sum, np.max, np.min, np.prod, np.mean]
        for fun in reduce_funcs:
            self.assertAlmostEqual(fun(v), fun(v.flatten()))

        other_funcs = [np.all, np.any, np.std, np.ptp]
        for fun in other_funcs:
            self.assertAlmostEqual(fun(v), fun(v.flatten()))

    def test_binary_ufuncs(self):
        v = BlockVector(2)
        v.set_blocks([np.ones(3) * 0.5, np.ones(2) * 0.8])
        v2 = BlockVector(2)
        v2.set_blocks([np.ones(3) * 3.0, np.ones(2) * 2.8])

        vi = BlockVector(2)
        vi.set_blocks([np.ones(3, dtype='i') * 5, np.ones(2, dtype='i') * 8])
        v2i = BlockVector(2)
        v2i.set_blocks([np.ones(3, dtype='i') * 3, np.ones(2, dtype='i') * 2])

        _int_ufuncs = {
            np.gcd,
            np.lcm,
            np.ldexp,
            np.left_shift,
            np.right_shift,
            np.bitwise_and,
            np.bitwise_or,
            np.bitwise_xor,
        }

        for fun in vec_binary_ufuncs:
            _v, _v2, _s = (vi, v2i, 3) if fun in _int_ufuncs else (v, v2, 3.0)

            flat_res = fun(_v.flatten(), _v2.flatten())
            res = fun(_v, _v2)
            self.assertTrue(np.allclose(flat_res, res.flatten()))

            res = fun(_v, _v2.flatten())
            self.assertTrue(np.allclose(flat_res, res.flatten()))

            res = fun(_v.flatten(), _v2)
            self.assertTrue(np.allclose(flat_res, res.flatten()))

            flat_res = fun(_v.flatten(), 5)
            res = fun(_v, 5)
            self.assertTrue(np.allclose(flat_res, res.flatten()))

            flat_res = fun(_s, _v2.flatten())
            res = fun(_s, _v2)
            self.assertTrue(np.allclose(flat_res, res.flatten()))

        v = BlockVector(2)
        a = np.ones(3, dtype=bool)
        b = np.ones(2, dtype=bool)
        v.set_block(0, a)
        v.set_block(1, b)

        v2 = BlockVector(2)
        a2 = np.zeros(3, dtype=bool)
        b2 = np.zeros(2, dtype=bool)
        v2.set_block(0, a2)
        v2.set_block(1, b2)

        binary_ufuncs = [np.logical_and, np.logical_or, np.logical_xor]
        for fun in binary_ufuncs:
            flat_res = fun(v.flatten(), v2.flatten())
            res = fun(v, v2)
            self.assertTrue(np.allclose(flat_res, res.flatten()))

    def test_min_with_empty_blocks(self):
        b = BlockVector(3)
        b.set_block(0, np.zeros(3))
        b.set_block(1, np.zeros(0))
        b.set_block(2, np.zeros(3))
        self.assertEqual(b.min(), 0)

    def test_max_with_empty_blocks(self):
        b = BlockVector(3)
        b.set_block(0, np.zeros(3))
        b.set_block(1, np.zeros(0))
        b.set_block(2, np.zeros(3))
        self.assertEqual(b.max(), 0)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"This is the deprecated pyomo.core.kernel.register_numpy_types module"

from pyomo.common.numeric_types import (
    RegisterNumericType,
    RegisterIntegerType,
    RegisterBooleanType,
    native_complex_types,
    native_numeric_types,
    native_integer_types,
    native_boolean_types,
)

from pyomo.common.dependencies import numpy, numpy_available as _has_numpy

# Ensure that the types were registered
bool(_has_numpy)

numpy_int_names = []
numpy_int = []
numpy_float_names = []
numpy_float = []
numpy_bool_names = []
numpy_bool = []
numpy_complex_names = []
numpy_complex = []

if _has_numpy:
    # Historically, the lists included several numpy aliases
    numpy_int_names.extend(('int_', 'intc', 'intp'))
    numpy_int.extend((numpy.int_, numpy.intc, numpy.intp))
    if hasattr(numpy, 'float_'):
        numpy_float_names.append('float_')
        numpy_float.append(numpy.float_)
    if hasattr(numpy, 'complex_'):
        numpy_complex_names.append('complex_')
        numpy_complex.append(numpy.complex_)

# Re-build the old numpy_* lists
for t in native_boolean_types:
    if t.__module__ == 'numpy':
        if t in native_integer_types:
            if t.__name__ not in numpy_int_names:
                numpy_int.append(t)
                numpy_int_names.append(t.__name__)
        elif t in native_numeric_types:
            if t.__name__ not in numpy_float_names:
                numpy_float.append(t)
                numpy_float_names.append(t.__name__)
        else:
            if t.__name__ not in numpy_bool_names:
                numpy_bool.append(t)
                numpy_bool_names.append(t.__name__)


# Complex
for t in native_complex_types:
    if t.__module__ == 'numpy':
        if t.__name__ not in numpy_complex_names:
            numpy_complex.append(t)
            numpy_complex_names.append(t.__name__)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.core.expr.taylor_series import taylor_series_expansion
from pyomo.core.expr import polynomial_degree, differentiate


class TestTaylorSeries(unittest.TestCase):
    def test_first_order_taylor_series(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.x.value = 1
        exprs_to_test = [m.x**2, pyo.exp(m.x), (m.x + 2) ** 2]
        for e in exprs_to_test:
            tsa = taylor_series_expansion(e)
            self.assertAlmostEqual(
                pyo.differentiate(e, wrt=m.x), pyo.differentiate(tsa, wrt=m.x)
            )
            self.assertAlmostEqual(pyo.value(e), pyo.value(tsa))
            self.assertEqual(polynomial_degree(tsa), 1)

    def test_higher_order_taylor_series(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=0.5)
        m.y = pyo.Var(initialize=1.5)

        e = m.x * m.y
        tse = taylor_series_expansion(
            e, diff_mode=differentiate.Modes.reverse_symbolic, order=2
        )
        for _x in [-2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2]:
            for _y in [-2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2]:
                m.x.value = _x
                m.y.value = _y
                self.assertAlmostEqual(pyo.value(e), pyo.value(tse))

        e = m.x**3 + m.y**3
        tse = taylor_series_expansion(
            e, diff_mode=differentiate.Modes.reverse_symbolic, order=3
        )
        for _x in [-2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2]:
            for _y in [-2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2]:
                m.x.value = _x
                m.y.value = _y
                self.assertAlmostEqual(pyo.value(e), pyo.value(tse))

        e = (m.x * m.y) ** 2
        tse = taylor_series_expansion(
            e, diff_mode=differentiate.Modes.reverse_symbolic, order=4
        )
        for _x in [-2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2]:
            for _y in [-2, -1.5, -1, -0.5, 0, 0.5, 1, 1.5, 2]:
                m.x.value = _x
                m.y.value = _y
                self.assertAlmostEqual(pyo.value(e), pyo.value(tse))
